/* Add an entry to a KDC, and extract the value to a local keytab 
 *
 * Simon Wilkinson 04/01/2001
 *
 * This would be a lot easier in Perl - and we really need some perl bindings
 * for the account creation stuff, if only to make automated account creation
 * easier. Hopefully from this it should be possible to build something
 * suitable using XS, and this will become obsolete
 *
 * Quick note - our error handling and cleanup in this sucks, big time. As
 * its a small application, it exit()s with any error, and doesn't bother
 * cleaning up on its way out. This is _not_ a good example to copy from
 * if you're planning to write anything long lived and/or interactive
 *
 * Another quick note - it might be (better, easier, quicker, faster) to
 * use kadmin -q to implement all of this - but it causes problems with 
 * having to enter passwords multiple times.
 */

#define KRB5_PRIVATE 1
 
#include <krb5.h>
#include <kadm5/admin.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <math.h>
#include <unistd.h>
#include <pwd.h>
#include <time.h>
#include <errno.h>

extern const struct _krb5_kt_ops krb5_ktf_writable_ops;

/* Globals suck */
void *handle = NULL;
krb5_context context;
krb5_boolean quiet=FALSE;

void init_kerberos(void) {
  krb5_error_code problem;
  
  if ((problem = krb5_init_context(&context))) {
    com_err("kdcregister", problem, "while initializing Kerberos");
    exit(1);
  }

  /* this produces an error message:
   * "kdcregister: Key table type is already registered. while registering key
   * table"
   * so we either don't need it anymore or more likely there is some problem
   * with having to rewrite the krb5_ktf_writable_ops declaration under 1.3.4?
   * normal behaviour still seems to work without this call
   */
  /*
  if ((problem = krb5_kt_register(context, &krb5_ktf_writable_ops))) {
    com_err("kdcregister", problem, "while registering key table");
    exit(1);
  }
  */

  {
    #define DEFAULT_KEYTAB "WRFILE:/etc/krb5.keytab"
    /* krb5_defkeyname is an internal library global and  should go away */
    extern char *krb5_defkeyname;
    krb5_defkeyname = DEFAULT_KEYTAB;
  }                                            
}

krb5_keytab resolve_keytab(char *keytab_str) {
  krb5_keytab keytab;
  krb5_error_code problem;
  
  if (keytab_str==NULL) {
    /* Using the default, that's nice and easy */  
    if ((problem=krb5_kt_default(context,&keytab))) {
      com_err("kdcregister", problem, "while opening keytab");
      exit(1);
    }
  } else {
    /* Not the default. Grrrr. We assume they've given us a proper filename
     * and not one with the kerberos action extension on it 
     */
    char *tmp = keytab_str;
   
    keytab_str = (char *) malloc(strlen("WRFILE:")+strlen(tmp)+1);
    if (keytab_str == NULL) {
      com_err("kdcregister", ENOMEM, "while creating keytab name");
      exit(1);
    }
    snprintf(keytab_str, strlen("WRFILE:")+strlen(tmp)+1, "WRFILE:%s", tmp);
    
    if ((problem = krb5_kt_resolve(context, keytab_str, &keytab))) {
      com_err("kdcregister", problem, "while resolving keytab %s", keytab_str);
      free(keytab_str);
      exit(1);
    }
    free(keytab_str);
  }
  return(keytab);
}

krb5_boolean principal_exists(krb5_principal principal, krb5_keytab keytab) {
  kadm5_ret_t problem;
  krb5_keytab_entry entry;
  
  problem = krb5_kt_get_entry(context, keytab, principal, 0, 0, &entry);
 
  if (problem && problem!=KRB5_KT_NOTFOUND && problem!=ENOENT) {
    com_err("kdcregister", problem, "while searching for principal in keytab");
    exit(1);
  }
  if (!problem) {
    if ((problem=krb5_kt_free_entry(context,&entry))) {
      com_err("kdcregister", problem, "while freeing keytab entry");
    }
  }
  return(problem==0);
}

void init_admin(char *princstr, char *keytab_str) {
  char *password = NULL;
  char *def_realm = NULL;
  char *newprincstr = princstr;
  
  kadm5_ret_t problem;
  krb5_ccache cc;
  krb5_principal princ;
  kadm5_config_params params;
  
  if ((problem = krb5_get_default_realm(context, &def_realm))) {
    com_err("kdcregister", problem, "while getting default realm");
    exit(1);
  }
  
  memset((char *) &params, 0, sizeof(params));
  
  params.mask |= KADM5_CONFIG_REALM;
  params.realm = def_realm;

  if ((problem = krb5_cc_default(context, &cc))) {
    com_err("kdcregister", problem, "while opening default ccache");
    exit(1);
  }
  
  if (princstr == NULL) {
    if (keytab_str != NULL) {
      if ((problem = krb5_sname_to_principal(context, NULL, "hostclient", 
  			                     KRB5_NT_SRV_HST, &princ))) {
        com_err("kdcregister", problem, "while creating host principal");
        exit(1);
      }
    } else {  				               
      if ((problem = krb5_cc_get_principal(context, cc, &princ))) {
        com_err("kdcregister", problem, "while getting default principal");
        exit(1);
      }
    }
    if ((problem = krb5_unparse_name(context, princ, &newprincstr))) {
      com_err("kdcregister", problem, "while canonicalizing principal");
      krb5_free_principal(context, princ);
      exit(1);
    }
    krb5_free_principal(context, princ);
  }
  
  if ((problem = kadm5_init_with_creds(newprincstr, cc, 
				      KADM5_ADMIN_SERVICE, &params,
				      KADM5_STRUCT_VERSION, 
				      KADM5_API_VERSION_2,
				      &handle))) {

/* com_err("kdcregister",problem,"while initialising with credentials"); */

    if ((problem = kadm5_init_with_skey(newprincstr, keytab_str,
    					KADM5_ADMIN_SERVICE, &params,
    					KADM5_STRUCT_VERSION,
    					KADM5_API_VERSION_2,
    					&handle))) {
      if (princstr==NULL) {
        com_err("kdcregister", problem, "while initialising with keytab");
        exit(1);
      }
      
      printf("Authenticating as %s with password.\n",princstr);
      problem=kadm5_init_with_password(princstr, password,
  				     KADM5_ADMIN_SERVICE, &params,
  				     KADM5_STRUCT_VERSION, KADM5_API_VERSION_2,
  				     &handle);
    }
  }

  if (problem) {
    com_err("kdcregister", problem, "while initialising kadmin interface");
    exit(1);
  }

  if ((problem = krb5_cc_close(context, cc))) {
    com_err("kdcregister", problem, "while closing ccache");
    exit(1);
  }

}

int quit()
{
  kadm5_destroy(handle);
  krb5_free_context(context);
  return 0;
}

/* This function adds a principal for the current host with a random key, and
 * then extracts that key to the default keytab
 */
 
int do_stuff(krb5_principal principal, krb5_keytab keytab, int force)
{
  kadm5_principal_ent_rec princ_rec;
  krb5_keytab_entry new_entry;
  krb5_keyblock *keys;
  kadm5_policy_ent_rec defpol;
  kadm5_ret_t problem;
  int nkeys, i;
  long mask = 0;
  char dummybuf[256];
  char *canon;
 
  /* Dummybuf is a scrambled password that we use when initially creating
   * the principal. We stop anyone from being able to use this, as we 
   * create the principal with NOTIX set */
    
  for (i = 0; i < 256; i++) 
    dummybuf[i]=(i+1) % 256;

  /* Zero all of the fields in the principal request structure */  
  memset((char *) &princ_rec, 0, sizeof(princ_rec));

  /********** Add the principal ***********/

  if ((problem = krb5_copy_principal(context, principal, &princ_rec.principal))) {
    com_err("kdcregister", problem, "while copying principal");
    exit(1);
  }
   
  if ((problem = krb5_unparse_name(context, princ_rec.principal, &canon))) {
    com_err("kdcregister", problem, "while canonicalizing principal");
    krb5_free_principal(context,princ_rec.principal);
    exit(1);
  }
  
  if (!kadm5_get_policy(handle, "default", &defpol)) {
    princ_rec.policy = "default";
    mask |= KADM5_POLICY;
    (void) kadm5_free_policy_ent(handle, &defpol);
  } else {
    if (!quiet) 
      fprintf(stderr,"WARNING: No policy specified, and no default policy\n");
  }
  mask &= ~KADM5_POLICY_CLR;
  mask |= KADM5_PRINCIPAL;
  
  /* We're going to use a random key, so we create the principal with a 
   * dummy password, and with ticket granting disabled */
     
  princ_rec.attributes |= KRB5_KDB_DISALLOW_ALL_TIX;
  mask |= KADM5_ATTRIBUTES;

  if ((problem = kadm5_create_principal(handle, &princ_rec, mask, dummybuf))) {
    if (problem != KADM5_DUP || !force) {
      com_err("kdcregister", problem, "while creating %s", canon);
      krb5_free_principal(context, princ_rec.principal);
      free(canon);
      exit(1);
    }
  }
  
  /* And now we create a random key - but get the keys back so that we
   * can stick them into the keytab file. We should perhaps be a bit more
   * careful with these arrays as they contain key material. */ 
  keys  = NULL;
  nkeys = 0;
  if ((problem = kadm5_randkey_principal(handle, princ_rec.principal, 
  	                                &keys, &nkeys))) {
    com_err("kdcregister", problem, " while randomizing key");
    krb5_free_principal(context, princ_rec.principal);
    free(canon);
    exit(1);
  }
  
  /* Having randomized the key, we now enable ticket granting */
  princ_rec.attributes &= ~KRB5_KDB_DISALLOW_ALL_TIX;
  mask = KADM5_ATTRIBUTES;
  if ((problem = kadm5_modify_principal(handle, &princ_rec, mask))) {
    com_err("kdcregister", problem, "while clearing DISALLOW_ALL_TIX");
    krb5_free_principal(context, princ_rec.principal);
    free(canon);
    exit(1);
  }
  
  if (!quiet) printf("Principal \"%s\" created.\n", canon);
  
  /******** Extract the principal to the key tab ************/
    
  /* We need to do a get principal to update the kvno stuff.
   */
  
  if ((problem = kadm5_get_principal(handle, princ_rec.principal, &princ_rec, 
                                    KADM5_PRINCIPAL_NORMAL_MASK))) {
    com_err("kdcregister", problem, "while getting principal");
    krb5_free_principal(context, princ_rec.principal);
    free(canon);
    exit(1); 
  }
  
  for (i = 0; i < nkeys; i++) {
    memset((char *) &new_entry, 0, sizeof(new_entry));
    new_entry.principal = princ_rec.principal;
    new_entry.key = keys[i];
    new_entry.vno = princ_rec.kvno;
    
    if ((problem = krb5_kt_add_entry(context, keytab, &new_entry))) {
      com_err("kdcregister", problem, "while adding key to keytab");
      kadm5_free_principal_ent(handle,&princ_rec);
      goto cleanup;
    }
  }
  kadm5_free_principal_ent(handle,&princ_rec);
  
cleanup:
  if (nkeys) {
    for (i = 0; i < nkeys; i++)
      krb5_free_keyblock_contents(context, &keys[i]);
    free(keys);
  }  
  
  return problem;
}

void usage () {
  fprintf(stderr,"kdcregister [-f] [-c] [-p principal] [-k source keytab] \\\n");  
  fprintf(stderr,"            [-t target keytab] principal\n");
}

int main(int argc, char *const *argv) {
  extern char *optarg;
  extern int optind;
  int optchar;
  char *princstr=NULL;
  char *source_keytab=NULL;
  char *target_keytab=NULL;
  krb5_principal create;
  krb5_boolean force=FALSE;
  krb5_boolean check=FALSE;
  krb5_keytab keytab;
  krb5_error_code problem;
  
  while ((optchar = getopt(argc, argv, "p:k:t:fhcq")) !=EOF) {
    switch(optchar) {
    case 'c':
      check=TRUE;
      break;
    case 'p':
      princstr = optarg;
      break;
    case 'k':
      source_keytab = optarg;
      break;
    case 't':
      target_keytab = optarg;
      break;
    case 'f':
      force=TRUE;
      break;
    case 'q':
      quiet=TRUE;
      break;
    case 'h':
      usage();
      exit(0);
      break;
    default:
      usage();
      exit(1);
      break;
    }
  }
  
  if (optind>=argc || argv[optind]==NULL) {
    usage();
    exit(1);
  }
  
  init_kerberos();
  keytab=resolve_keytab(target_keytab);

  if ((problem=krb5_parse_name(context,argv[optind],&create))) {
    com_err("kdcregister", problem, "while parsing principal");
    exit(1);
  }
  
  /* We want to check to see if the principal already exists in the given
   * keytab - if it does we don't bother recreating it */
  
  if (check && principal_exists(create,keytab)) {
      if (!quiet) printf("Principal already exists in keytab\n");
      exit(0);
  }

  init_admin(princstr,source_keytab);
  do_stuff(create,keytab, force);
  quit();
  
  exit(0);
}
